﻿-----------------------------------------------------------------------------
--
-- data
--   \deaths
--     \<deathName>
--       - count
--
-- For non-grouped units:
--       \deathTimes (Array)
--
-- For grouped units:
--       \deathList (array)
--			- time
--			\ recentEventList
--				{ timeOffset, health, reportedHealth, healthChange, maximumHealth, details, message, type, typeLocal, abilityName }
--
-----------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("AssessmentDeaths")
L:RegisterTranslations("enUS", function() return {
	["Deaths List"] = true,
	["Deaths"] = true,
	["Death Times"] = true,
	["Kill List"] = true,
	["Kill Times"] = true,

	["Death"] = true,
	["Death Options."] = true,

	["Melee"] = true,
	
	["Death Recent Event Log"] = true,
	
	["Group Only"] = true,
	["Only record kills caused by group members."] = true,

	["Death Event Log History Time"] = true,
	["Select the amount of time to buffer events to display leading up to a group member's death.\nUse 0 to disable event log recording."] = true,
	["Predict Event Log Health"] = true,
	["Predict player's actual health for the death event log.\nDisabling this option will show the health as reported by the API, which can be lagged."] = true,
	
	["Death Log Health Change"] = true,
	
	["Common Health"] = true,
	["Damage"] = true,
	["Heal"] = true,

	["Colors"] = true,
	["Color options"] = true,
	["Common Health"] = true,
	["Color for unchanged section of health in death event logs."] = true,
	["Damage"] = true,
	["Color for damage sections of health in death event logs."] = true,
	["Healing"] = true,
	["Color for heal sections of health in death event logs."] = true,

} end)

L:RegisterTranslations("koKR", function() return {
	["Deaths List"] = "죽음 목록",
	["Deaths"] = "죽음",
	["Death Times"] = "죽은 횟수",
	["Kill List"] = "죽임 목록",
	["Kill Times"] = "죽임 횟수",

	["Death"] = "죽음",
	["Death Options."] = "죽음 설정입니다.",
	["Group Only"] = "파티만",
	["Only record kills caused by group members."] = "파티원에 의한 죽임만 기록합니다.",
} end)

-----------------------------------------------------------------------------

local AssessmentDeaths = Assessment:NewModule("Deaths", "AceEvent-2.0", "Parser-3.0")
AssessmentDeaths.db = Assessment:AcquireDBNamespace("Deaths")
Assessment:RegisterDefaults("Deaths", "profile", {
	enable				= true,
	gatherGroupOnly		= false,
	eventLogTime		= 10,

	predictHealth		= true,
	
	commonHealthColor	= { 0.4, 0.4, 0.4, 1 },
	damageColor			= { 1, 0, 0, 1 },
	healColor			= { 0, 1, 0, 1 },
})

local profile
local emptyMethod = Assessment.EmptyMethod
local new, del, reset, copy = Assessment.NewTable, Assessment.ReleaseTable, Assessment.ResetTable, Assessment.CopyTable
local pairs = pairs
local time, GetTime = time, GetTime
local UnitHealth, UnitHealthMax = UnitHealth, UnitHealthMax

local meleeText = L["Melee"]

-----------------------------------------------------------------------------

Assessment.TagData.Deaths =
{
	method = function(data)
				local fightData = data.fightData
				if not fightData then return 0 end
				
				local deaths = fightData.deaths
				if not deaths then return 0 end
				
				local playerDeaths = deaths[data.key]
				if not playerDeaths then return 0 end

				return playerDeaths.count
			 end,
	data = "deaths",
	guaranteeResult = true,
	numeric = true,
}

-----------------------------------------------------------------------------

Assessment.ListType.DeathsList =
{
	method = function(fightData)
				if not fightData or not fightData.deaths then return emptyMethod end
				return pairs(fightData.deaths)
			 end,
	data = "deaths",
	description = L["Deaths List"],
}

local commonHealthIteratorText = L["Common Health"]
local damageIteratorText = L["Damage"]
local healIteratorText = L["Heal"]
local function DeathLogHealthChangeIterator(state, index)
	if not index then
		local health = state.health
		local healthChange = state.healthChange
		if healthChange < 0 then
			health = health + healthChange
			if health < 0 then
				health = 0
			end
		end
		return commonHealthIteratorText, health
	elseif index == commonHealthIteratorText then
		local damage = -state.healthChange
		if damage <= 0 then
			damage = 0
		else
			local health = state.health
			if damage > health then
				damage = health
			end
		end
		return damageIteratorText, damage
	elseif index == damageIteratorText then
		local heal = state.healthChange
		if heal <= 0 then
			heal = 0
		else
			local maxHeal = state.maximumHealth - state.health
			if heal > maxHeal then
				heal = maxHeal
			end
		end
		return healIteratorText, heal
	end
end

Assessment.ListType.DeathLogHealthChange =
{
	method = function(fightData, ...)
				local data = fightData.deaths
				if not data then return emptyMethod end

				for i=1,select('#',...) do
					if not data then return emptyMethod end
					data = data[select(i, ...)]
				end
				if not data then return emptyMethod end
				
				return DeathLogHealthChangeIterator, data, nil
			 end,

	data = "deaths",
	description = L["Death Log Health Change"]
}

-----------------------------------------------------------------------------

local keyToColorNameMap = { [L["Common Health"]] = "commonHealthColor", [L["Damage"]] = "damageColor", [L["Heal"]] = "healColor" }
Assessment.ColorType.DeathLogHealthChange = 
{
	method = function(fightData, data)
				local colorName = keyToColorNameMap[data.key]
				if not colorName then return nil end
				return profile[colorName]
			 end,
	description = L["Death Log Health Change"]
}

-----------------------------------------------------------------------------

Assessment.ListType.EnumerateTable.updateData.deaths = "deaths"

-----------------------------------------------------------------------------

Assessment.DefaultDisplaySets.DeathLogHealthChange =
{
	titleTag			= "[WindowParameter(1)]",
	listType			= "DeathLogHealthChange", 
	listDataTag			= "[WindowParameterList]",
	barValueTag			= "Value", 
	filterTag			= "1",
	fightIndex			= -1,
	text				= { { }, { tag= "[BarValue:Short]" } },
	description			= L["Death Log Health Change"],
	colorTypeList		= { "DeathLogHealthChange", count = 1 },
	sortKeyTag			= "Rank",
	descendingSort		= false,
	favorite			= false,
	selectable			= false,
}

Assessment.DefaultDisplaySets.DeathRecentEventLog =
{
	titleTag			= "Death #[WindowParameter(2)], [WindowParameter(1)]",
	listType			= "EnumerateTable",
	listDataTag			= "deaths,[WindowParameter(1)],deathList,[WindowParameter(2)],recentEventList",
	barValueTag			= "(Maximum(GetValueField(\"healthChange\"), 0) + GetValueField(\"health\"))/GetValueField(\"maximumHealth\")", 
	normalizeBarValues	= false,
	sortKeyTag			= "GetValueField(\"timeOffset\")",
	descendingSort		= false,
	fightIndex			= -1,
	text				= { count = 4, 
							{ tag = "[GetValueField(\"timeOffset\"):Format(\"%.1f\")]s", alignment = "RIGHT", offsets = { left = 0, right = 21 } },
							{ tag = "[GetValueField(\"type\") == \"Fade\":\"- \"][GetValueField(\"type\") == \"Aura\":\"+ \"][GetValueField(\"abilityName\") || GetValueField(\"typeLocal\"):Fuchsia][GetValueField(\"sourceName\"):HideIf(1):Surround(\" (\", \")\")]", alignment = "LEFT", offsets = { left = 22, right = -25 } },
							{ tag = "[GetValueField(\"health\"):HideIf(GetValueField(\"healthChange\")==0):Suffix(\"/\")][GetValueField(\"maximumHealth\"):HideIf(GetValueField(\"healthChange\")==0)] [(100*GetValueField(\"health\")/GetValueField(\"maximumHealth\")):HideIf(GetValueField(\"healthChange\")==0):HideZero:Percent:Bracket]", visible = false, alignment = "RIGHT", offsets = { left = 22, right = -25 }  },
							{ tag = "[GetValueField(\"details\") || Select(GetValueField(\"healthChange\") > 0, (\"+\"..GetValueField(\"healthChange\")):Green, GetValueField(\"healthChange\"):HideZero:Red)]", alignment = "RIGHT", offsets = { left = 0, right = 0 } },
						  },
	description			= L["Death Recent Event Log"],
	defaultColor		= { 0.5, 0.5, 0.5 },
	selectable			= false,
	segmentEnable		= true,
	segmentDisplaySet	= L["Death Log Health Change"],
	segmentData			= "[WindowParameter(1)],deathList,[WindowParameter(2)],recentEventList,[Key]",
}

Assessment.DefaultDisplaySets.DeathTimes =
{
	titleTag			= "[WindowParameter(1)]",
	listType			= "EnumerateTable", 
	listDataTag			= "deaths,[WindowParameter(1)],deathList",
	barValueTag			= "0", 
	filterTag			= "1",
	fightIndex			= -1,
	text				= { count = 2, { tag= "Death [Rank]" }, { tag = "[GetValueField(\"time\"):FormatTime]" } },
	description			= L["Death Times"],
	selectable			= false,
	defaultColor		= { 0.8, 0.8, 0.8 },
	onClickType			= "SHOW_DISPLAY_SET",
	onClickDisplaySet	= L["Death Recent Event Log"],
	onClickData			= "[WindowParameter(1)],[Key]",
}

Assessment.DefaultDisplaySets.Deaths =
{
	titleTag			= "Deaths",
	listType			= "GroupMembers", 
	barValueTag			= "Deaths", 
	filterTag			= "BarValue ~= 0",
	fightIndex			= 0,
	description			= L["Deaths"],
	onClickType			= "SHOW_DISPLAY_SET",
	onClickDisplaySet	= L["Death Times"],
	onClickData			= "[Key]",
	defaultColor		= { 0.8, 0.8, 0.8 },
	colorTypeList		= { "Class", count = 1 },
}

Assessment.DefaultDisplaySets.KillTimes =
{
	titleTag			= "[WindowParameter(1)]",
	listType			= "EnumerateTable", 
	listDataTag			= "deaths,[WindowParameter(1)],deathTimes",
	barValueTag			= "0", 
	filterTag			= "1",
	fightIndex			= -1,
	text				= { count = 2, { tag= "Kill [Rank]" }, { tag = "[Value:FormatTime]" } },
	description			= L["Kill Times"],
	selectable			= false,
	defaultColor		= { 0.8, 0.3, 0.5 },
}

Assessment.DefaultDisplaySets.KillList =
{
	titleTag			= "Kill List",
	listType			= "DeathsList", 
	barValueTag			= "Deaths", 
	filterTag			= "~IsGroupMember",
	fightIndex			= 0,
	description			= L["Kill List"],
	onClickType			= "SHOW_DISPLAY_SET",
	onClickDisplaySet	= L["Kill Times"],
	onClickData			= "[Key]",
	defaultColor		= { 0.8, 0.3, 0.5 },
	colorTypeList		= { "Class", count = 1 },
}

-----------------------------------------------------------------------------

Assessment.Options.args.data.args.death =
{
	type = "group",
	name = L["Death"],
	desc = L["Death Options."],
	args =
	{
		groupOnly =
		{
			type = "toggle",
			name = L["Group Only"],
			desc = L["Only record kills caused by group members."],
			get = function() return profile.gatherGroupOnly end,
			set = function(v) profile.gatherGroupOnly = v end,
		},
		eventLogTime =
		{
			type = "range",
			name = L["Death Event Log History Time"],
			desc = L["Select the amount of time to buffer events to display leading up to a group member's death.\nUse 0 to disable event log recording."],
			min = 0,
			max = 30,
			step = 0.5,
			get = function() return profile.eventLogTime end,
			set = function(v) profile.eventLogTime = v end
		},
		predictHealth =
		{
			type = "toggle",
			name = L["Predict Event Log Health"],
			desc = L["Predict player's actual health for the death event log.\nDisabling this option will show the health as reported by the API, which can be lagged."],
			get = function() return profile.predictHealth end,
			set = function(v) profile.predictHealth = v end,
		},
		color =
		{
			type = "group",
			name = L["Colors"],
			desc = L["Color options"],
			args =
			{
				commonHealth =
				{
					type = "color",
					name = L["Common Health"],
					desc = L["Color for unchanged section of health in death event logs."],
					get = function() return unpack(profile.commonHealthColor) end,
					set = function(r, g, b, a) profile.commonHealthColor = { r, g, b, a } Assessment:RedrawAll() end,
					hasAlpha = true
				},
				damage =
				{
					type = "color",
					name = L["Damage"],
					desc = L["Color for damage sections of health in death event logs."],
					get = function() return unpack(profile.damageColor) end,
					set = function(r, g, b, a) profile.damageColor = { r, g, b, a } Assessment:RedrawAll() end,
					hasAlpha = true
				},
				healing  =
				{
					type = "color",
					name = L["Healing"],
					desc = L["Color for heal sections of health in death event logs."],
					get = function() return unpack(profile.healColor) end,
					set = function(r, g, b, a) profile.healColor = { r, g, b, a } Assessment:RedrawAll() end,
					hasAlpha = true
				},
			},
		},
	}
}

-----------------------------------------------------------------------------

Assessment.DefaultTags.Deaths =
{
	tag			= "[Deaths]",
	fightIndex	= 0,
	description	= L["Deaths"],
}

-----------------------------------------------------------------------------

function AssessmentDeaths:OnInitialize()
	profile = self.db.profile
end

function AssessmentDeaths:OnEnable()
	self:RegisterParserEvent({ eventType = "Death" }, "OnParserDeathEvent")
	
	-- For event logs
	self:RegisterParserEvent({ eventType = "Damage", recipientID_not = false }, "OnParserDamageEvent")
	self:RegisterParserEvent({ eventType = "Heal", recipientID_not = false }, "OnParserHealEvent")
	self:RegisterParserEvent({ eventType = "Aura", recipientID_not = false }, "OnParserAuraEvent")
	self:RegisterParserEvent({ eventType = "Fade", recipientID_not = false }, "OnParserFadeEvent")
	self:RegisterParserEvent({ eventType = "Miss", recipientID_not = false }, "OnParserMissEvent")
	self:RegisterParserEvent({ eventType = "Environmental", recipientID_not = false }, "OnParserEnvironmentalEvent")

	self:RegisterEvent("Assessment:StartNewFight", "OnNewFight")
end

function AssessmentDeaths:OnDisable()
end

-----------------------------------------------------------------------------

local eventLogHistory = { }
local tinsert = table.insert

function AssessmentDeaths:OnNewFight()
	reset(eventLogHistory, true)
end

local function ProcessLogEvent(eventData, healthChange, abilityName, details)
	local unitid = eventData.recipientID
	
	local recipientName = eventData.recipientName
	local unitHistory = eventLogHistory[recipientName]
	if not unitHistory then
		unitHistory = new()
		eventLogHistory[recipientName] = unitHistory
	end

	local time = GetTime()
	local eventRecord = new()
	local health = UnitHealth(unitid)
	local maximumHealth = UnitHealthMax(unitid)
	
	eventRecord.reportedHealth	= health
	
	if profile.predictHealth then
		if health == unitHistory.health then
			health = unitHistory.predictedHealth
			unitHistory.previousHealth = health
		elseif health == unitHistory.previousHealth then
			unitHistory.health = health
			health = unitHistory.predictedHealth
			unitHistory.previousHealth = health
		else
			unitHistory.health = health
			unitHistory.previousHealth = nil
		end

		local predictedHealth = health + healthChange
		if predictedHealth <= 0 then
			unitHistory.predictedHealth = 0
		elseif predictedHealth >= maximumHealth then
			unitHistory.predictedHealth = maximumHealth
		else
			unitHistory.predictedHealth = predictedHealth
		end		
	end

	eventRecord.time			= time
	eventRecord.sourceName		= eventData.sourceName
	eventRecord.health			= health
	eventRecord.healthChange	= healthChange
	eventRecord.details			= details
	eventRecord.maximumHealth	= maximumHealth
	eventRecord.type			= eventData.eventType
	eventRecord.typeLocal		= eventData.eventTypeLocal
	eventRecord.abilityName		= abilityName
	eventRecord.message			= eventData.message

	tinsert(unitHistory, 1, eventRecord)

	local eraseTime = time-profile.eventLogTime
	
	for i=#unitHistory, 1, -1 do
		local history = unitHistory[i]
		if history.time > eraseTime then
			return
		end
		unitHistory[i] = del(history)
	end
end

function AssessmentDeaths:OnParserDamageEvent(eventData)
	ProcessLogEvent(eventData, -eventData.amount, eventData.abilityName or meleeText)
end

function AssessmentDeaths:OnParserHealEvent(eventData)
	ProcessLogEvent(eventData, eventData.amount, eventData.abilityName)
end

function AssessmentDeaths:OnParserAuraEvent(eventData)
	if eventData.applications ~= 1 then
		ProcessLogEvent(eventData, 0, eventData.abilityName.." ("..eventData.applications..")")
	else
		ProcessLogEvent(eventData, 0, eventData.abilityName)
	end
end

function AssessmentDeaths:OnParserFadeEvent(eventData)
	ProcessLogEvent(eventData, 0, eventData.abilityName)
end

function AssessmentDeaths:OnParserMissEvent(eventData)
	ProcessLogEvent(eventData, 0, eventData.abilityName or meleeText, eventData.missTypeLocal)
end

function AssessmentDeaths:OnParserEnvironmentalEvent(eventData)
	ProcessLogEvent(eventData, -eventData.amount, eventData.hazardTypeLocal)
end

function AssessmentDeaths:OnGroupMemberDeath(recipientName, deathIndex)
	local copyFrom = eventLogHistory[recipientName]
	if not copyFrom then return end
	
	for data in self:ForceIterateUpdateDataTables() do
		local deaths = data.deaths
		if deaths then
			local recipientDeath = deaths[recipientName]
			if recipientDeath then
				local deathList = recipientDeath.deathList
				if deathList then
					deathList = deathList[#deathList]
					if deathList and not deathList.recentEventList then
						local recentEventList = new()
						local referenceTime = deathList.referenceTime
						
						if referenceTime then
							for i = #copyFrom, 1, -1 do
								local e = copy(copyFrom[i])
								e.timeOffset = e.time - referenceTime
								recentEventList[#recentEventList+1] = e
							end
						end
						
						deathList.recentEventList = recentEventList
					end
				end
			end
		end
	end
	
	eventLogHistory[recipientName] = del(copyFrom)
end

-----------------------------------------------------------------------------

function AssessmentDeaths:OnParserDeathEvent(eventData)
	local recipientID = eventData.recipientID 
	if recipientID then
		-- It's a group member that died.. record it.
		ProcessLogEvent(eventData, 0)
	else
		-- A non-group member..
		
		-- If the options say to only gather kills done by group members, check that.
		if profile.gatherGroupOnly then
			-- If it is group only, then check that the killer was in our group
			if not eventData.sourceID then return end
		else
			-- The option is to gather all kills. In this case, exit if there is a source name, to stop slain messages from counting twice
			if eventData.sourceName then return end
		end
	end

	local deathTime = time()
	local time = GetTime()

	for data in self:ForceIterateUpdateDataTables() do
		data.dataUpdateTimes.deaths = time
		local deaths = data.deaths
		if not deaths then
			deaths = new()
			data.deaths = deaths
		end

		local recipientName = eventData.recipientName
		local recipientDeath = deaths[recipientName]
		if recipientDeath then
			recipientDeath.count = recipientDeath.count + 1
		else
			recipientDeath = new()
			recipientDeath.count = 1
			deaths[recipientName] = recipientDeath
		end

		if recipientID then
			local deathList = recipientDeath.deathList
			if not deathList then
				deathList = new()
				
				local deathTimes = recipientDeath.deathTimes
				if deathTimes then
					-- Promote deathTimes array to deathList, and remove deathTimes
					for i=1,#deathTimes do
						local t = new()
						t.time = deathTime
						deathList[i] = t
					end
					
					recipientDeath.deathTimes = del(deathTimes)
				end
			
				recipientDeath.deathList = deathList
			end

			local record = new()
			record.time = deathTime
			record.referenceTime = time
			deathList[#deathList+1] = record
		
			if profile.eventLogTime > 0 then
				self:ScheduleEvent(self.OnGroupMemberDeath, 0.1, self, eventData.recipientName)
			end
		else
			local deathList = recipientDeath.deathList
			if deathList then
				local record = new()
				record.time = deathTime
				deathList[#deathList+1] = record
			else
				local deathTimes = recipientDeath.deathTimes
				if not deathTimes then
					deathTimes = new()
					recipientDeath.deathTimes = deathTimes
				end
				
				deathTimes[#deathTimes+1] = deathTime
			end		
		end
	end
end

-----------------------------------------------------------------------------
